@@ -1,29 +1,33 @@
-# ZestyParser 0.7.0 -- Parses in Python zestily
+# ZestyParser 0.8.0 -- Parses in Python zestily
 # Copyright (C) 2006-2007 Adam Atlas
 # 
-# This program is free software; you can redistribute it and/or
-# modify it under the terms of the GNU General Public License
-# as published by the Free Software Foundation; either version 2
-# of the License, or (at your option) any later version.
+# Permission is hereby granted, free of charge, to any person obtaining a copy
+# of this software and associated documentation files (the "Software"), to deal
+# in the Software without restriction, including without limitation the rights
+# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+# copies of the Software, and to permit persons to whom the Software is
+# furnished to do so, subject to the following conditions:
 # 
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
+# The above copyright notice and this permission notice shall be included in
+# all copies or substantial portions of the Software.
 # 
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+# THE SOFTWARE.
 
 '''
-@group Basic Tokens: RawToken,Token,TakeToken
+@group Basic Tokens: Raw,RE,RawToken,Token,TakeToken
 @group Complex Tokens: CompositeToken,TokenSequence,TokenSeries
 @group Special Tokens: Defer,Default,Lookahead,Negative
 @group TokenSequence Flags: Omit,Skip,Only
 
-@version: 0.7.0
+@version: 0.8.0
 @author: Adam Atlas
-@copyright: Copyright 2006-2007 Adam Atlas. Released under the terms of the GNU General Public License.
+@copyright: Copyright 2006-2007 Adam Atlas. Released under the MIT license (see LICENSE.txt).
 @contact: adam@atlas.st
 
 @var EmptyToken: A L{Default} instance initialized with the empty string.
@@ -34,19 +38,20 @@ In ZestyParser, a token object must, at minimum, be a callable taking a L{ZestyP
 The L{Tokens} module contains a variety of predefined token classes (instances of which are callable) and other valid token objects which should cover most parsing situations.
 '''
 
-import re, copy, sys, types, new
+import re, copy, types, warnings
 from Parser import NotMatched, ParseError
 
-__all__ = ('AbstractToken', 'TokenWrapper', 'Token', 'RawToken', 'CompositeToken', 'TokenSequence', 'TakeToken', 'TokenSeries', 'EmptyToken', 'Default', 'Skip', 'Omit', 'Only', 'Defer', 'Lookahead', 'Negative', 'EOF', 'Const')
+__all__ = ('Placeholder', 'AbstractToken', 'TokenWrapper', 'RE', 'Raw', 'Token', 'RawToken', 'CompositeToken', 'TokenSequence', 'TakeToken', 'TokenSeries', 'EmptyToken', 'Default', 'Skip', 'Omit', 'Only', 'Defer', 'Lookahead', 'Negative', 'EOF', 'Whitespace', 'Const', 'Inf')
+
 
 rstack = []
+replstack = []
+
+Inf = -1
 
 def count_args(callable):
     t = type(callable)
-    if t in (types.TypeType, types.BuiltinFunctionType):
-        #builtin methods don't provide argcounts, so the best we can do is assume they only take the data itself as a parameter
-        return 1
-    elif t is types.FunctionType:
+    if t is types.FunctionType:
         return callable.func_code.co_argcount
     elif t is types.ClassType:
         return callable.__init__.im_func.func_code.co_argcount - 1
@@ -54,8 +59,55 @@ def count_args(callable):
         return callable.__call__.im_func.func_code.co_argcount - 1
     elif t is types.MethodType:
         return callable.im_func.func_code.co_argcount - 1
+    #assume it's some builtin that only takes the data itself as a parameter
+    return 1
 
-class AbstractToken:
+class Placeholder:
+    def __init__(self, key=None):
+        self.key = key
+    
+    def __eq__(self, key):
+        return self.key == key
+    
+    def __hash__(self):
+        return hash(self.key)
+    
+    def __call__(self, key=None):
+        return Placeholder(key)
+    
+    def _single_replace(cls, subj, vals, kwvals):
+        if isinstance(subj, cls):
+            if vals and subj == None:
+                return vals.pop(0)
+            elif subj in kwvals and subj != None:
+                return kwvals[subj]
+        elif isinstance(subj, AbstractToken):
+            subj._replace(vals, kwvals)
+        return subj
+    _single_replace = classmethod(_single_replace)
+    
+    def _list_replace(cls, subj, vals, kwvals):
+        for i, v in enumerate(subj):
+            subj[i] = cls._single_replace(v, vals, kwvals)
+    _list_replace = classmethod(_list_replace)
+    
+    def __repr__(self):
+        return '<Placeholder %r>' % self.key
+
+class ListReplacing:    
+    def _replace(self, vals, kwvals):
+        if self not in replstack:
+            replstack.append(self)
+            Placeholder._list_replace(self.desc, vals, kwvals)
+            replstack.pop()
+class SingleReplacing:        
+    def _replace(self, vals, kwvals):
+        if self not in replstack:
+            replstack.append(self)
+            self.desc = Placeholder._single_replace(self.desc, vals, kwvals)
+            replstack.pop()
+
+class AbstractToken (object):
     '''
     Base class from which most tokens defined in this module derive. Subclassing this is not required for writing tokens, since they can be any callable with certain semantics, but this class provides several useful services for creating reusable token classes, such as callback support and convenient operator overloading.
     
@@ -64,18 +116,18 @@ class AbstractToken:
     @ivar to: An optional callable which, if not None, will be called in the same manner as a callback (after any callback and before returning to the parser instance), but will be passed only one argument: the data matched (or returned by the callback, if any). Its main purpose is to allow you to concisely do things like C{Token('[0-9]+', group=0, to=int)} -- the builtin callable C{int} will be passed the text matched by the regex, so the token will ultimately return an integer instead of a string or a regex match object. You can also use this property with L{AHT} types, for more complex multi-stage parsing. See the C{n3.py} and C{n3rdflib.py} examples for a demonstration of this. (In previous versions, this was passed to the initializer as C{as}, but this is deprecated because C{as} will become a reserved word in Python 2.6. Change your code to use {to}.)
     '''
     
+    name = None
     failMessage = None
-    callbackArgCount = False
-    _call_orig = None
     callback = None
     to = None
 
     #'as' is deprecated in favor of 'to' since it's becoming a reserved word
     def __init__(self, desc, callback=None, to=None, as=None, name=None):
-        self._call_orig = self.__call__
         self.desc = desc
         self.callback = callback
-        self.to = self.as = to or as
+        self.to = to or as
+        if as:
+            warnings.warn('`as` argument is deprecated; use `to`', DeprecationWarning, stacklevel=2)
         self.name = name
     
     def __repr__(self):
@@ -83,42 +135,98 @@ class AbstractToken:
 
     def __str__(self): return ''
 
-    def _call_preprocess(self, parser, origCursor):
-        data = self._call_orig(parser, origCursor)
-        if self.callback is not None:
-            if self.callbackArgCount == 2:
-                data = self.callback(parser, data)
-            elif self.callbackArgCount == 1:
-                data = self.callback(data)
-            elif self.callbackArgCount == 3:
-                data = self.callback(parser, data, origCursor)
-        if self.to is not None: data = self.to(data)
-        return data
+    def _make_callbackrun(self, func, callback):
+        argcount = count_args(callback)
+        if argcount == 1:
+            def f(parser, origCursor):
+                return callback(func(parser, origCursor))
+        elif argcount == 2:
+            def f(parser, origCursor):
+                return callback(parser, func(parser, origCursor))
+        elif argcount == 3:
+            def f(parser, origCursor):
+                return callback(parser, func(parser, origCursor), origCursor)
+        return f
     
-    def _call_failcheck(self, parser, origCursor):
-        try:
-            data = self._call_preprocess(parser, origCursor)
-            return data
-        except NotMatched:
-            if self.failMessage: raise ParseError(parser, self.failMessage)
-            else: raise
+    def _make_torun(self, func):
+        def f(parser, origCursor):
+            return self.to(func(parser, origCursor))
+        return f
+    
+    def _make_failcheck(self, func):
+        def f(parser, origCursor):
+            try:
+                data = func(parser, origCursor)
+                return data
+            except NotMatched:
+                if self.failMessage is True:
+                    raise ParseError(parser, 'Expected %s' % str(self))
+                elif self.failMessage:
+                    raise ParseError(parser, self.failMessage)
+                else: raise
+        return f
+    
+    def _poke(self):
+        c = self.__call__
+        if self.callback:
+            c = self._make_callbackrun(c, self.callback)
+        if self.to:
+            c = self._make_torun(c)
+        if self.failMessage:
+            c = self._make_failcheck(c)
+        
+        if c is self.__call__ and not isinstance(self, Defer):
+            self.parse = None
+            del self.parse
+        else:
+            self.parse = c
         
     def __copy__(self):
-        n = new.instance(self.__class__)
+        n = self.__class__.__new__(self.__class__)
         n.__dict__.update(self.__dict__)
-        if '__call__' in n.__dict__:
-            del n.__call__
-        n._call_orig = n.__call__
+        n._poke()
+        n.desc = copy.copy(self.desc)
         return n
+    
+    def __setattr__(self, name, value):
+        super(AbstractToken, self).__setattr__(name, value)
+        if name in ('callback', 'failMessage', 'to'):
+            self._poke()
 
     def __add__(self, other):
         '''Allows you to construct L{TokenSequence}s with the + operator.'''
         return TokenSequence([self, other])
     
+    def __sub__(self, other):
+        '''Allows you to construct L{TokenSequence}s with the - operator, automatically padded with L{Whitespace}.
+        
+I realize it's a bit weird to use the - operator for this, but the main motivation is giving it the same precedence as +. Still, you can read it as a sort of "blank" (which is what the left and right tokens are being joined by), instead of "minus".'''
+        return TokenSequence([self, Whitespace, other])
+    
     def __or__(self, other):
         '''Allows you to construct L{CompositeToken}s with the | operator.'''
         return CompositeToken([self, other])
     
+    def __mul__(self, val):
+        '''Allows you to construct L{TokenSeries} with the * operator. Operand can be:
+        
+            - int (a series of exactly this many)
+            - (int, ) (a series of at least this many)
+            - (x:int, y:int) a series of x to y
+        
+        The constant Inf can be used in some of these -- * Inf yields a 0--infinity series, and * (x, Inf) yields an x--infinity series.
+        '''
+        t = TokenSeries(self)
+        if isinstance(val, int):
+            t.min = t.max = val
+        elif isinstance(val, tuple):
+            if len(val) == 2:
+                t.min, t.max = val
+            elif len(val) == 1:
+                t.min = val
+        return t
+    __rmul__ = __mul__
+    
     def __rshift__(self, callback):
         '''
         Convenience overloading for setting the L{callback<AbstractToken.callback>} of a token whose initializer you do not call directly, such as the result of combining tokens with L{+<__add__>} or L{|<__or__>}.
@@ -145,29 +253,37 @@ class AbstractToken:
     
     def __invert__(self):
         return Negative(self)
+        
+    def _replace(self, vals, kwvals):
+        pass
     
-    def __setattr__(self, name, value):
-        self.__dict__[name] = value
-        if name in ('callback', 'failMessage', 'to'):
-            if self.failMessage:
-                self.__call__ = self._call_failcheck
-            elif self.callback or self.to:
-                self.__call__ = self._call_preprocess
-            else:
-                self.__call__ = self._call_orig
-        if name == 'callback':
-            self.callbackArgCount = count_args(value)
+    def __imod__(self, val):
+        if isinstance(val, (tuple, list)):
+            self._replace(val, {})
+        elif isinstance(val, dict):
+            self._replace([], val)
+        else:
+            self._replace([val], {})
+        return self
+    
+    def __mod__(self, val):
+        new = copy.copy(self)
+        new %= val
+        return new
 
 class TokenWrapper (AbstractToken):
     '''If you write your own token type in a way other than subclassing AbstractToken, e.g. by simply writing a function, you can use this as a decorator to automatically let it take advantage of AbstractToken's magic.
     '''
     def __call__(self, parser, origCursor):
-        return self.desc(parser, origCursor)
+        t = parser.scan(self.desc)
+        if parser.last:
+            return t
+        else: raise NotMatched
 
     def __str__(self):
         return repr(self.desc)
 
-class Token (AbstractToken):
+class RE (AbstractToken):
     '''
     A class whose instances match Python regular expressions.
     
@@ -183,7 +299,12 @@ class Token (AbstractToken):
         '''
         if not hasattr(regex, 'match'):
             regex = re.compile(regex, re.DOTALL)
-        AbstractToken.__init__(self, regex, **kwargs)
+        super(Token, self).__init__(regex, **kwargs)
+        if group is not None:
+            try:
+                group = int(group)
+            except ValueError:
+                raise ValueError('got non-numeric value for `group`: ' + group)
         self.group = group
     
     def __call__(self, parser, origCursor):
@@ -198,7 +319,9 @@ class Token (AbstractToken):
     def __str__(self):
         return repr(self.desc.pattern)
 
-class RawToken (AbstractToken):
+Token = RE
+
+class Raw (AbstractToken):
     '''
     A class whose instances match only a particular string. Returns that string.
     
@@ -212,7 +335,7 @@ class RawToken (AbstractToken):
         @param caseInsensitive: To be set as the object's L{caseInsensitive} property.
         @type caseInsensitive: bool
         '''
-        AbstractToken.__init__(self, string, **kwargs)
+        super(Raw, self).__init__(string, **kwargs)
         self.len = len(string)
         self.caseInsensitive = caseInsensitive
         if caseInsensitive:
@@ -229,6 +352,8 @@ class RawToken (AbstractToken):
     def __str__(self):
         return repr(self.desc)
 
+RawToken = Raw
+
 class Default (AbstractToken):
     '''
     A class whose instances always return L{desc} and do not advance the parser's cursor.
@@ -240,12 +365,12 @@ class Default (AbstractToken):
 
 EmptyToken = Default('')
 
-class CompositeToken (AbstractToken):
+class CompositeToken (ListReplacing, AbstractToken):
     '''
     A class whose instances match any of a number of tokens.
     
-    @ivar desc: An iterable returning token objects.
-    @type desc: iterable
+    @ivar desc: A list of token objects.
+    @type desc: list
     '''
     def __call__(self, parser, origCursor):
         for t in self.desc:
@@ -276,7 +401,7 @@ class CompositeToken (AbstractToken):
             self.desc.append(other)
         return self
 
-class TokenSequence (AbstractToken):
+class TokenSequence (ListReplacing, AbstractToken):
     '''
     A class whose instances match a sequence of tokens. Returns a corresponding list of return values from L{ZestyParser.scan}.
     
@@ -284,14 +409,17 @@ class TokenSequence (AbstractToken):
     
     If the sequence contains an L{Only} token, its result will be returned instead of the usual list, though it still requires that subsequent tokens match. Multiple L{Only} tokens are meaningless and L{TokenSequence}'s behavior in that case is undefined. 
     
-    @ivar desc: An iterable returning token objects.
-    @type desc: iterable
+    @ivar desc: A list of token objects.
+    @type desc: list
     '''
+    
     def __call__(self, parser, origCursor):
         o = []
         only = False
         onlyVal = None
         for g in self.desc:
+            if g is Whitespace:
+                parser.skip(parser.whitespace)
             r = parser.scan(g)
             if parser.last is None:
                 raise NotMatched
@@ -299,7 +427,7 @@ class TokenSequence (AbstractToken):
                 only = True
                 onlyVal = r
                 continue
-            if not isinstance(g, (Skip, Omit)) and not only:
+            if not isinstance(g, (Skip, Omit, _Whitespace)) and not only:
                 o.append(r)
         if only: #heh
             return onlyVal
@@ -321,6 +449,12 @@ class TokenSequence (AbstractToken):
         else:
             return TokenSequence(self.desc + [other])
     
+    def __sub__(self, other):
+        if hasattr(other, '__iter__'):
+            return TokenSequence(self.desc + [Whitespace] + list(other))
+        else:
+            return TokenSequence(self.desc + [Whitespace, other])
+    
     def __iadd__(self, other):
         if hasattr(other, '__iter__'):
             self.desc += list(other)
@@ -334,7 +468,7 @@ class TakeToken (AbstractToken):
     '''
     
     def __init__(self, length, **kwargs):
-        AbstractToken.__init__(self, length, **kwargs)
+        super(TakeToken, self).__init__(length, **kwargs)
     
     def __call__(self, parser, start):
         end = start + self.desc
@@ -342,7 +476,7 @@ class TakeToken (AbstractToken):
         parser.cursor = end
         return parser.data[start:end]
 
-class TokenSeries (AbstractToken):
+class TokenSeries (SingleReplacing, AbstractToken):
     '''
     A particularly versatile class whose instances match one token multiple times (with a great degree of customizability).
     
@@ -371,7 +505,7 @@ class TokenSeries (AbstractToken):
     @ivar until: An optional 2-tuple whose first item is a token, and whose second item is either a message or False. The presence of this property indicates that the token in C{until[0]} must match at the end of the series. If this fails, then if C{until[1]} is a message, a ParseError will be raised with that message; if it is False, NotMatched will be raised.
     '''
     def __init__(self, token, min=0, max=-1, skip=None, prefix=None, postfix=None, delimiter=None, until=None, includeDelimiter=False, **kwargs):
-        AbstractToken.__init__(self, token, **kwargs)
+        super(TokenSeries, self).__init__(token, **kwargs)
         self.min, self.max, self.skip, self.prefix, self.postfix, self.delimiter, self.until, self.includeDelimiter = min, max, skip, prefix, postfix, delimiter, until, includeDelimiter
 
     def __call__(self, parser, origCursor):
@@ -401,7 +535,10 @@ class TokenSeries (AbstractToken):
         if self.until and not done:
                 parser.cursor = c
                 if self.until[1]:
-                    raise ParseError(parser, self.until[1])
+                    if self.until[1] is True:
+                        raise ParseError(parser, 'Expected %s' % self.until[0])
+                    else:
+                        raise ParseError(parser, self.until[1])
                 else:
                     raise NotMatched
         if i >= self.min:
@@ -417,41 +554,42 @@ class Defer (AbstractToken):
     A token which takes a callable (generally a lambda) which takes no arguments and itself returns a token. A Defer instance, upon being called, will call this function, scan for the returned token, and return that return value. This is primarily intended to allow you to define tokens recursively; if you need to refer to a token that hasn't been defined yet, simply use C{Defer(lambda: T_SOME_TOKEN)}, where C{T_SOME_TOKEN} is the token's eventual name.
     '''
     
-    def __init__(self, func, **kwargs):
-        AbstractToken.__init__(self, func, **kwargs)
-    
     def __call__(self, parser, origCursor):
-        t = parser.scan(self.desc())
-        if parser.last is None: raise NotMatched
-        return t
+        self.__call__ = self.desc()
+        self._poke()
+        return self.parse(parser, origCursor)
 
-class Skip (AbstractToken):
+def _pad(p, subj, outer=True):
+    if isinstance(subj, TokenSequence):
+        t = copy.copy(subj)
+        t.desc = list(t.desc)
+        start = (not outer) and 1 or 0
+        stop = 2*(len(t.desc) + (outer and 1 or -1))
+        for i in range(start, stop, 2): t.desc.insert(i, p)
+        return t
+    else:
+        return TokenSequence([p, Only(subj), p])
+
+class Skip (SingleReplacing, AbstractToken):
     '''
     See L{TokenSequence}.
     '''
     def __call__(self, parser, origCursor):
         parser.skip(self.desc)
     
-    def pad(self, tok):
+    def pad(self, tok, outer=True):
         '''
-        Takes a TokenSequence and returns a copy with this L{Skip} token separating every element therein.
+        Takes a TokenSequence and returns a copy with this L{Skip} token separating every element therein. Alternately, takes any other token and returns a TokenSequence equivalent to (self + Only(tok) + self).
         
         @param tok: The token sequence to operate on.
-        @type tok: L{TokenSequence}
+        @type tok: L{TokenSequence} or other token
+        @param outer: If operating on a L{TokenSequence}, whether to also include self on either side of the sequence, in addition to between each element.
+        @type outer: bool
         @return: L{TokenSequence}
         '''
-        t = copy.copy(tok)
-        o = []
-        i = 0
-        while i < len(t.desc):
-            o.append(t.desc[i])
-            if i != len(t.desc)-1:
-                o.append(self)
-            i += 1
-        t.desc = o
-        return t
+        return _pad(self, tok, outer)
 
-class Omit (AbstractToken):
+class Omit (SingleReplacing, AbstractToken):
     '''
     See L{TokenSequence}.
     '''
@@ -459,12 +597,12 @@ class Omit (AbstractToken):
         if not parser.skip(self.desc):
             raise NotMatched
 
-class Only (TokenWrapper):
+class Only (SingleReplacing, TokenWrapper):
     '''
     See L{TokenSequence}.
     '''
 
-class Lookahead (AbstractToken):
+class Lookahead (SingleReplacing, AbstractToken):
     '''
     Scans for another token and returns its result as usual, but doesn't actually advance the parser's cursor.
     '''
@@ -474,7 +612,7 @@ class Lookahead (AbstractToken):
         if parser.last is None: raise NotMatched
         return t
 
-class Negative (AbstractToken):
+class Negative (SingleReplacing, AbstractToken):
     '''
     Scans for another token, and only matches (returning True) if that token did not match.
     '''
@@ -484,7 +622,7 @@ class Negative (AbstractToken):
         if parser.last: raise NotMatched
         return True
 
-class _EOF(AbstractToken):
+class _EOF (AbstractToken):
     '''
     Matches returning None if the parser is at the end of its input.
     '''
@@ -492,6 +630,11 @@ class _EOF(AbstractToken):
         if parser.cursor != parser.len: raise NotMatched
 EOF = _EOF(None)
 
+class _Whitespace (AbstractToken):
+    def __call__(self, parser, origCursor):
+        parser.skip(parser.whitespace)
+Whitespace = _Whitespace(None)
+
 def Const(value):
     '''
     This is a factory function used for when you want an L{AbstractToken}-compatible callback that returns a constant. Example::
