@@ -1,10 +1,10 @@
 from ZestyParser import *
 from sys import stdin
 
-T_HELLO = Token('(?i)hello', group=0)
-T_SP = Token('\s+', group=0)
-T_WORLD = Token('(?i)world', group=0)
-T_EXCL = RawToken('!')
+T_HELLO = RE('(?i)hello', group=0)
+T_SP = RE('\s+', group=0)
+T_WORLD = RE('(?i)world', group=0)
+T_EXCL = Raw('!')
 
 #p = ZestyParser('hello world WORLD !! hello !')
 p = ZestyParser(stdin.read())
